from sh import infrable

from infrable import __version__


def test_remote():
    out = infrable.remote("--help")
    assert "-" in out
    assert "affected-hosts" in out
    assert "infra-hosts" in out
    assert "dev_host" in out
    assert "beta_host" in out
    assert "prod_host" in out
    assert "managed_hosts" in out



import shutil
from pathlib import Path

from sh import infrable


def pytest_configure(config):
    Path("infra.py").unlink(missing_ok=True)
    shutil.rmtree("templates", ignore_errors=True)
    shutil.rmtree("modules", ignore_errors=True)
    infrable.init()
    infrable.env.dev()



__version__ = "0.1.42"


from pathlib import Path

INFRA_MODULE_NAME = "infra"
IS_PROJECT_DIR = Path(f"{INFRA_MODULE_NAME}.py").exists()

if IS_PROJECT_DIR:
    from infrable.host import Host
    from infrable.meta import Meta
    from infrable.readfile import readfile
    from infrable.service import Service
    from infrable.switch import Switch
    from infrable.utils import concurrent, concurrentcontext



from pathlib import Path

from sh import infrable

from infrable import __version__

data = Path("tests/data")


def gen_test_data():
    infrable.hosts(_out=data.joinpath("hosts"))
    infrable.hosts(repr=True, _out=data.joinpath("hosts_repr"))
    infrable.hosts(format="{name},{ip},{fqdn}", _out=data.joinpath("hosts_format"))


def test_hosts():
    assert infrable.hosts(_tty_out=False) == data.joinpath("hosts").read_text()
    assert (
        infrable.hosts(repr=True, _tty_out=False)
        == data.joinpath("hosts_repr").read_text()
    )
    assert (
        infrable.hosts(format="{name},{ip},{fqdn}", _tty_out=False)
        == data.joinpath("hosts_format").read_text()
    )



from sh import infrable

from infrable import __version__


def test_switch():
    infrable.switch.env.beta()
    assert infrable.switch.env(_tty_out=False).strip() == "beta"
    assert infrable.switches(_tty_out=False).strip() == "env = beta"

    infrable.switch.env.prod()
    assert infrable.switch.env(_tty_out=False).strip() == "prod"
    assert infrable.switches(_tty_out=False).strip() == "env = prod"

    infrable.env.dev()
    assert infrable.switch.env(_tty_out=False).strip() == "dev"
    assert infrable.switches(_tty_out=False).strip() == "env = dev"



from pathlib import Path

from sh import infrable

from infrable import __version__

data = Path("tests/data")


def gen_test_data():
    infrable.services(_out=data.joinpath("services"))
    infrable.services(repr=True, _out=data.joinpath("services_repr"))
    infrable.services(
        format="{name},{host.ip},{meta}", _out=data.joinpath("services_format")
    )


def test_services():
    assert infrable.services(_tty_out=False) == data.joinpath("services").read_text()
    assert (
        "nginx = Service(host=Host(ip='127.0.0.1"
        in data.joinpath("services_repr").read_text()
    )
    assert (
        infrable.services(format="{name},{host.ip},{meta}", _tty_out=False)
        == data.joinpath("services_format").read_text()
    )



from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Callable, Generator, Iterable, Iterator, TypeVar

from jinja2 import Template

T1 = TypeVar("T1")
T2 = TypeVar("T2")


def item_formatter(template: str) -> Template:
    return Template(
        template,
        variable_start_string="{",
        variable_end_string="}",
        autoescape=False,
        keep_trailing_newline=True,
        finalize=lambda x: x or "",
    )


@contextmanager
def concurrentcontext(
    function: Callable[[T1], T2],
    generator: Iterable[T1],
    *,
    workers: int | None = None,
) -> Generator[Iterator[T2], None, None]:
    """With context, run a function on a batch of arguments concurrently."""

    with ThreadPoolExecutor(max_workers=workers) as executor:
        yield executor.map(function, generator)


def concurrent(
    function: Callable[[T1], T2],
    generator: Iterable[T1],
    *,
    workers: int | None = None,
) -> list[T2]:
    """Run a functions on a batch of arguments in concurrently."""

    with concurrentcontext(function, generator, workers=workers) as results:
        return list(results)



from infrable import Host, Service, infra


def test_items():
    assert len(infra.items) == 27
    assert infra.items["dev"] == "dev"


def test_item_types():
    assert len(infra.item_types) == 11
    assert infra.item_types[Host]["dev_host"].fqdn == "dev.example.com"
    assert infra.item_types[Service]["nginx"].host.fqdn == "dev.example.com"


def test_hosts():
    assert len(infra.hosts) == 3
    assert infra.hosts["dev_host"].fqdn == "dev.example.com"


def test_services():
    assert len(infra.services) == 2
    assert infra.services["nginx"].host.fqdn == "dev.example.com"


def test_host_groups():
    assert len(infra.host_groups) == 1
    assert infra.host_groups["managed_hosts"][0].fqdn == "dev.example.com"


def test_typers():
    assert len(infra.typers) == 3
    assert infra.typers["nginx"].registered_commands[0].name == "reload"


def test_metas():
    assert len(infra.metas) == 0


def test_by_fqdn():
    assert infra.by_fqdn["dev.example.com"].fqdn == "dev.example.com"


def test_by_ip():
    assert infra.by_ip["127.0.0.1"].fqdn == "dev.example.com"


def test_get_host():
    assert infra.get_host("dev_host").fqdn == "dev.example.com"
    assert infra.get_host("dev.example.com").fqdn == "dev.example.com"
    assert infra.get_host("127.0.0.1").fqdn == "dev.example.com"
    assert infra.get_host("foo") == None



from typing import Iterable, Optional

import typer
from tqdm import tqdm

from infrable import Host, files, infra, utils

app = typer.Typer(no_args_is_help=True)


@app.command(name="-")
def from_stdin(command: str, workers: Optional[int] = None):
    """Execute a script on the list of hosts passed via stdin."""

    hosts = []
    for line in typer.get_text_stream("stdin").readlines():
        host = line.split(maxsplit=1)[0].split("@", 1)[-1].strip()
        if not host:
            continue
        for host in infra.filtered_hosts(only=[host]):
            hosts.append(host)

    _execute_batch(hosts, command=command, workers=workers)


@app.command()
def infra_hosts(
    command: str,
    only: Optional[list[str]] = None,
    workers: Optional[int] = None,
):
    """Execute a script on the hosts listed in the infra."""

    hosts = infra.hosts.values()
    if only:
        hosts = infra.filtered_hosts(only=only)

    _execute_batch(hosts, command=command, workers=workers)


@app.command()
def affected_hosts(
    command: str,
    only: Optional[list[str]] = None,
    workers: Optional[int] = None,
):
    """Execute a script on the affected hosts in the last deployment."""

    hosts = files.affected_hosts()

    if only:
        only_hosts = set(h.fqdn for h in infra.filtered_hosts(only=only))
        hosts = (h for h in hosts if h.fqdn in only_hosts)

    _execute_batch(hosts, command=command, workers=workers)


def _create_host_groups_command(group):
    def main(
        command: str, only: Optional[list[str]] = None, workers: Optional[int] = None
    ):
        hosts = group
        if only:
            only_hosts = set(h.fqdn for h in infra.filtered_hosts(only=only))
            hosts = (h for h in hosts if h.fqdn in only_hosts)

        _execute_batch(hosts, command=command, workers=workers)

    return main


for name, group in infra.host_groups.items():
    help = f"Execute a script on {name} hosts."
    cmd = _create_host_groups_command(group)
    app.command(name=name, help=help)(cmd)


def _create_host_command(host):
    def main(command: str):
        host.remote()(command, _fg=True)

    return main


for name, host in infra.hosts.items():

    help = f"Execute a script on {name}."
    cmd = _create_host_command(host)
    app.command(name=name, help=help)(cmd)


def _execute_batch(hosts: Iterable[Host], command: str, workers: int | None):
    hosts = {h.ip: h for h in hosts}.values()
    fn = lambda host: (host, host.remote()(command, _err_to_out=True))
    with utils.concurrentcontext(fn, hosts, workers=workers) as results:
        for host, result in tqdm(results, total=len(hosts)):
            typer.secho(f"╭ {host}", bold=True)
            print(result)
            print()



import typer
from click import Choice

from infrable import infra

app = typer.Typer(no_args_is_help=True)


def _create_switch_command(switch):
    def main(
        value: str = typer.Argument(None, click_type=Choice(list(switch.options))),
        options: bool = False,
    ):
        if options:
            for opt in sorted(switch.options):
                print(opt)
            return

        if value is None:
            print(switch())
        else:
            switch.set(value)

    return main


for name, switch in infra.switches.items():
    help = f"Get or set the value of the {name} switch."
    cmd = _create_switch_command(switch)
    app.command(name=name, help=help)(cmd)



import os
import shutil
from pathlib import Path

import pytest
from sh import infrable

from infrable import __version__, paths

data = Path("tests/data")


def gen_test_data():
    infrable.files.gen(_out=data.joinpath("pre_files_gen"))
    infrable.files.gen("templates/nginx", _out=data.joinpath("files_gen_nginx"))
    infrable.files.gen(
        "templates/nginx/web.j2", _out=data.joinpath("files_gen_nginx_web")
    )
    infrable.files.gen(_out=data.joinpath("post_files_gen"))
    _gen_old_files(copy=False)
    infrable.files.diff(_out=data.joinpath("files_diff"))


def _gen_old_files(copy: bool):
    for new in paths.files.glob("**/*.new"):
        old = (paths.files / new.relative_to(paths.files)).with_suffix(".old")
        text = new.read_text() if copy else ""
        old.write_text(text)


def test_files_gen():
    shutil.rmtree(paths.files, ignore_errors=True)
    assert set(infrable.files.gen(_tty_out=False).splitlines()) == set(
        data.joinpath("pre_files_gen").read_text().splitlines()
    )
    assert set(
        infrable.files.gen("templates/nginx", _tty_out=False).splitlines()
    ) == set(data.joinpath("files_gen_nginx").read_text().splitlines())
    assert set(
        infrable.files.gen("templates/nginx/web.j2", _tty_out=False).splitlines()
    ) == set(data.joinpath("files_gen_nginx_web").read_text().splitlines())
    assert set(infrable.files.gen(_tty_out=False).splitlines()) == set(
        data.joinpath("post_files_gen").read_text().splitlines()
    )


def test_files_diff():
    shutil.rmtree(paths.files, ignore_errors=True)
    infrable.files.gen()
    _gen_old_files(copy=True)
    assert infrable.files.diff(_tty_out=False) == ""
    _gen_old_files(copy=False)
    assert set(infrable.files.diff(_tty_out=False).splitlines()) == set(
        data.joinpath("files_diff").read_text().splitlines()
    )


def test_files_backup():
    infrable.files.gen()
    _gen_old_files(copy=False)
    backups_count = len(list(paths.backups.glob("*")))
    out = infrable.files.backup(_tty_out=False)
    assert out.strip().startswith(f"backup: {paths.backups}/")
    assert len(list(paths.backups.glob("*"))) == backups_count + 1


def test_files_revert():
    infrable.files.gen()
    _gen_old_files(copy=False)
    infrable.files.backup()

    news1, olds1 = [f.read_text() for f in paths.files.glob("**/*.new")], [
        f.read_text() for f in paths.files.glob("**/*.old")
    ]

    out = infrable.files.revert(_tty_out=False)
    assert out.strip().startswith(f"reverting from: {paths.backups}/")

    news2, olds2 = [f.read_text() for f in paths.files.glob("**/*.new")], [
        f.read_text() for f in paths.files.glob("**/*.old")
    ]

    assert news1 == olds2
    assert olds1 == news2

    bkpdir = out.split(": ", 1)[-1].strip()
    assert infrable.files.revert(bkpdir, _tty_out=False) == out


def test_files_affect_hosts():
    infrable.files.gen()
    _gen_old_files(copy=True)
    assert infrable.files("affected-hosts", _tty_out=False) == ""
    _gen_old_files(copy=False)
    assert (
        infrable.files("affected-hosts", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="dev_host", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="dev.example.com", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="127.0.0.1", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="managed_hosts", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="prod_host", _tty_out=False).strip() == ""
    )


def test_symlinks():
    os.symlink("../tests/data", paths.templates / "symlink")
    try:
        with pytest.raises(Exception) as e:
            infrable.files.gen()
    except Exception as e:
        os.remove(paths.templates / "symlink")
        raise e


def test_hidden_files():
    hiddendir = paths.templates.joinpath(".config")
    hiddendir.mkdir(parents=True, exist_ok=True)
    hiddendir.joinpath(".hiddenfile").touch(exist_ok=True)

    try:
        with pytest.raises(Exception) as e:
            infrable.files.gen()
    except Exception as e:
        shutil.rmtree(hiddendir, ignore_errors=True)
        raise e



import os
import sys
from typing import Any, Iterable, Type

from typer import Typer

from infrable import INFRA_MODULE_NAME, errors
from infrable.host import Host
from infrable.meta import Meta
from infrable.service import Service
from infrable.switch import Switch

items: dict[str, Any] = {}
item_types: dict[Type[Any], dict[str, Any]] = {}

hosts: dict[str, Host] = {}
services: dict[str, Service] = {}
host_groups: dict[str, list[Host]] = {}
switches: dict[str, Switch] = {}
typers: dict[str, Typer] = {}
metas: dict[str, Meta] = {}

by_fqdn: dict[str, Host] = {}
by_ip: dict[str, Host] = {}


if not items:
    if not os.path.exists(f"{INFRA_MODULE_NAME}.py"):
        raise errors.InfraDotPyNotFoundError(INFRA_MODULE_NAME)

    sys.path.append(".")
    infra_module = __import__(INFRA_MODULE_NAME)

    for name in dir(infra_module):
        if name.startswith("_"):
            # Skip private members
            continue

        if name in items:
            raise errors.DuplicateNameError(name)
        item = getattr(infra_module, name)
        items[name] = item
        itemtype = type(item)
        if itemtype not in item_types:
            item_types[itemtype] = {}
        item_types[itemtype][name] = item

        bases = list(itemtype.__bases__)
        while bases:
            base = bases.pop()
            if base != object:
                if base not in item_types:
                    item_types[base] = {}
                item_types[base][name] = item
                bases.extend(base.__bases__)

        if isinstance(item, Host):
            hosts[name] = item
            if item.ip in by_ip:
                raise errors.DuplicateIPError(item.ip)
            by_ip[item.ip] = item
            if item.fqdn:
                if item.fqdn in by_fqdn:
                    raise errors.DuplicateFQDNError(item.fqdn)
                by_fqdn[item.fqdn] = item
            if item.typer:
                typers[name] = item.typer
        elif isinstance(item, Service):
            services[name] = item
            if item.typer:
                typers[name] = item.typer
        elif isinstance(item, Switch):
            switches[name] = item
            if item.typer:
                typers[name] = item.typer
        elif isinstance(item, Meta):
            metas[name] = item
            if "typer" in item:
                if isinstance(item, Typer):
                    typers[name] = item.typer
        elif isinstance(item, Typer):
            typers[name] = item
        elif isinstance(item, list) and item and isinstance(item[0], Host):
            host_groups[name] = item
        elif t := getattr(item, "typer", None):
            if isinstance(t, Typer):
                typers[name] = t


def get_host(name: str) -> Host | None:
    """Get a host by name, FQDN, IP or service name."""

    if host := hosts.get(name) or by_fqdn.get(name) or by_ip.get(name):
        return host
    if host := services.get(name):
        return host.host
    return None


def filtered_hosts(only: Iterable[str] | None = None) -> list[Host]:
    """Filter hosts by name, FQDN, IP, service name or host group."""

    if not only:
        return list(hosts.values())

    filtered = []
    for o in only or []:
        if item := get_host(o):
            filtered.append(item)
        elif item := host_groups.get(o):
            for h in item:
                filtered.append(h)
        else:
            raise errors.HostOrServiceNotFoundError(o)

    return filtered



from pathlib import Path
from typing import Optional

import typer

from infrable import files

app = typer.Typer(no_args_is_help=True)


@app.command(help=files.affected_hosts.__doc__)
def affected_hosts(only: Optional[list[str]] = None):
    for host in files.affected_hosts(only=only):
        print(host)


@app.command(help=files.deploy.__doc__)
def deploy(
    path: Optional[Path] = typer.Argument(None),
    only: Optional[list[str]] = None,
    yes: bool = False,
    workers: Optional[int] = None,
):
    files.deploy(path, only=only, yes=yes, workers=workers)


@app.command(help=files.recover.__doc__)
def recover(
    path: Optional[Path] = typer.Argument(None),
    yes: bool = False,
    workers: Optional[int] = None,
):
    files.recover(path, yes=yes, workers=workers)


@app.command(help=files.gen.__doc__)
def gen(
    path: Optional[Path] = typer.Argument(None),
    only: Optional[list[str]] = None,
):
    files.gen(path, only=only)


@app.command(help=files.backup.__doc__)
def backup():
    files.backup()


@app.command(help=files.pull.__doc__)
def pull(workers: Optional[int] = None):
    files.pull(workers=workers)


@app.command(help=files.diff.__doc__)
def diff():
    files.diff()


@app.command(help=files.push.__doc__)
def push(yes: bool = False, workers: Optional[int] = None):
    files.push(yes=yes, workers=workers)


@app.command(help=files.revert.__doc__)
def revert(path: Path = typer.Argument(None)):
    files.revert(path)



import shutil
from pathlib import Path

from sh import infrable

from infrable import __version__

data = Path("tests/data")


def gen_test_data():
    Path("infra.py").unlink(missing_ok=True)
    shutil.rmtree("templates", ignore_errors=True)
    shutil.rmtree("modules", ignore_errors=True)
    infrable.init(_out=data.joinpath("pre_init"))
    infrable.init(_out=data.joinpath("post_init"))


def test_version():
    assert infrable.version().strip() == f"infrable {__version__}"


def test_init():
    Path("infra.py").unlink(missing_ok=True)
    shutil.rmtree("modules", ignore_errors=True)
    shutil.rmtree("templates", ignore_errors=True)
    assert infrable.init() == data.joinpath("pre_init").read_text()
    assert infrable.init() == data.joinpath("post_init").read_text()
